import os
from setuptools import setup, find_packages, Command, Extension


__version__ = None
exec(open('fanc/version.py').read())


class CleanCommand(Command):
    """
    Custom clean command to tidy up the project root.
    """
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('rm -vrf ./build ./dist ./*.pyc ./*.tgz ./*.egg-info ./htmlcov')


setup(
    name='fanc',
    version=__version__,
    description='Framework for the ANalysis of C-data.',
    setup_requires=[
        'setuptools>=18.0',
        'cython'
    ],
    packages=find_packages(),
    package_data={'fanc': ['test/data/*/*']},
    install_requires=[
        'numpy>=1.16.0',
        'scipy',
        'pillow',
        'matplotlib',
        'pandas>=0.15.0',
        'pysam>=0.9.1',
        'biopython',
        'pytest',
        'msgpack',
        'msgpack-numpy>=0.4.4.3',
        'scikit-learn',
        'progressbar2',
        'pybedtools',
        'pyBigWig',
        'PyYAML>=5.1',
        'tables>=3.5.1',
        'seaborn',
        'future',
        'gridmap>=0.14.0',
        'intervaltree',
        'genomic_regions>=0.0.7',
        'scikit-image>=0.15.0',
        'cooler>=0.8.0',
        'h5py'
    ],
    scripts=['bin/fanc', 'bin/fancplot'],
    cmdclass={
        'clean': CleanCommand
    },
    ext_modules=[
        Extension(
            'fanc.tools.sambam',
            sources=['fanc/tools/sambam.pyx', 'fanc/tools/natural_cmp.c'],
        ),
    ],
)



"""

FAN-C
=====

Provides
    1. Classes for working with Hi-C data
    2. Classes for working with tabular data

"""
import logging
import os

from genomic_regions import GenomicRegion
from .architecture.domains import InsulationScore, InsulationScores, DirectionalityIndex
from .architecture.compartments import ABCompartmentMatrix
from .architecture.comparisons import FoldChangeMatrix, DifferenceMatrix
from .architecture.aggregate import AggregateMatrix, aggregate_boundaries, aggregate_loops
from .config import config
from .general import FileBased
from .map import *
from .hic import Hic
from .matrix import Edge, RegionMatrix
from .pairs import ReadPairs
from .peaks import RaoPeakInfo
from .regions import Genome, Chromosome
from .registry import class_id_dict
from .version import __version__
from .tools.load import load

# configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

example_data = dict(
    hic="test/data/test_network/rao2014.chr11_77400000_78600000.hic",
    chip_bigwig="test/data/test_plotting/CTCF_ChIP_FE_chr11_77-80Mb_mouse_embryo_fibroblasts.bigwig",
    chip_bedgraph="test/data/test_plotting/CTCF_ChIP_FE_chr11_77-80Mb_mouse_embryo_fibroblasts.bedgraph.gz",
    chip_peak_bed="test/data/test_plotting/CTCF_ChIP_FE_chr11_77-80Mb_mouse_embryo_fibroblasts.peaks.bed.gz",
    gene_gtf="test/data/test_plotting/genes_mm10_chr11_77-80Mb.gtf.gz",
)
_basepath = os.path.abspath(os.path.dirname(__file__))
example_data = {k: os.path.join(_basepath, v) for k, v in example_data.items()}



"""
Module that handles FAN-C configuration files.

Generally, some FAN-C behaviour can be controlled using a "fanc.conf"
file in one of the following locations, in order of preference:

1. fanc.conf in current directory
2. File specified by environment variable fanc_CONF
3. fanc.conf in home folder
4. .fanc.conf in home folder
5. ~/.config/fanc.conf
6. /etc/fanc/fanc.conf

All configuration files will be read, but the options of lower-ranking
locations will be overwritten.

To write a default configuration file use  :code:`fanc write-config`
from the command line or

.. code::

    from fanc.config import write_default_config

    write_default_config('~/.fanc.conf')  # or any other location

"""

from fanc.tools.general import Map
import yaml
import os
import copy
import warnings
import logging
logger = logging.getLogger(__name__)

_default_config_content = """\
#
# MEMORY
#
edge_buffer_size: 3G

#
# HDF5
#
hdf5_compression_level: 1
hdf5_compression_algorithm: blosc

#
# Mapping
#
bowtie2_path: bowtie2
bwa_path: bwa

#
# PLOTTING
#
colormap_hic: germany
adjustment_slider_height: .2
pad_empty_axis: .1
pad_with_label: .2
pad_with_ticks: .1
pad_next_title: .2
pad_with_tick_legend: .1

#
# EMAIL
#
email_from_address:
email_smtp_server:
email_smtp_port:
email_smtp_username:
email_smtp_password:

#
# ERROR HANDLING
#
trap_signals:
    - SIGUSR1
    - SIGUSR2
raise_exception_on_trapped_signal: True

#
# Progressbars
#
hide_progressbars: False

#
# Sun/Oracle Grid Engine
#
gridmap_tmpdir:

sge_qsub_path: qsub
sge_qdel_path: qdel
sge_default_queue: all.q
sge_log_dir:
sge_parallel_environment: smp
sge_nodes:
sge_qsub_options: "-V -notify"
sge_shell: /bin/bash

#
# SLURM
#

slurm_sbatch_path: sbatch
slurm_sbatch_options: ""
slurm_scancel_path: scancel
slurm_log_dir:
slurm_shell: /bin/bash

#
# Juicer
#
juicer_tools_jar_path: 
"""
default_config = yaml.load(_default_config_content, Loader=yaml.FullLoader)

_config_file_name = 'fanc.conf'
_environment_variable_name = 'FANC_CONF'

# check common places for config file
# modified from: http://stackoverflow.com/questions/7567642/where-to-put-a-configuration-file-in-python
# the order of files in this tuple determines config file priority
# options in files with higher priority override options in files with lower priority
# default options have lowest priority
_config_file_locations = (
    # current directory
    os.path.join(os.curdir, _config_file_name),
    # environment variable
    os.environ.get(_environment_variable_name),
    # home folder
    os.path.join(os.path.expanduser("~"), _config_file_name),
    # home folder (hidden file)
    os.path.join(os.path.expanduser("~"), '.' + _config_file_name),
    # home folder (.config directory)
    os.path.join(os.path.expanduser("~"), '.config', 'fanc', _config_file_name),
    # /etc/fanc (MySQL style)
    os.path.join("/etc/fanc/", _config_file_name)
)


# read the first config file that is found, checking in the order above
def read_file_configs(config_file_locations=None):
    """
    Read all fanc.conf configuration files found in
    1. current directory
    2. specified by environment variable fanc_CONF
    3. fanc.conf in home folder
    4. .fanc.conf in home folder
    5. ~/.config/fanc.conf
    6. /etc/fanc/fanc.conf

    :param config_file_locations: optional custom config file paths in a list
    :return: dictionaries of parsed configs
    """

    if config_file_locations is None:
        config_file_locations = _config_file_locations

    configs = []
    for location in config_file_locations:
        if location is None:
            continue

        config_path = os.path.expanduser(location)
        if os.path.exists(config_path):
            logger.debug("Loading config from {}".format(config_path))
            try:
                with open(config_path, 'r') as config_file:
                    config_file_content = config_file.read()
                    file_config = yaml.safe_load(config_file_content)
                    configs.append((config_path, file_config))
            except IOError as e:
                logger.error("Could not read config file {}".format(config_path), e)
                pass
    return configs


file_configs = read_file_configs()

_config_dict = copy.deepcopy(default_config)
for current_config_path, current_file_config in reversed(file_configs):
    for key, value in current_file_config.items():
        if key not in default_config:
            warnings.warn("Config option '{}' is not known".format(key))
        else:
            _config_dict[key] = value

config = Map(_config_dict)


def write_default_config(file_name=None, overwrite=False):
    """
    Write the default configuration file.

    :param file_name: Output path, e.g. ~/.fanc.conf
    :param overwrite: Overwrite existing file if True. Default: False
    """
    if file_name is None:
        file_name = os.path.expanduser('~/.fanc.conf')
    else:
        file_name = os.path.expanduser(file_name)
    if not overwrite and os.path.exists(file_name):
        raise IOError("Config file exists in this location ({}). "
                      "Use overwrite=True or remove file.".format(file_name))

    with open(file_name, 'w') as output_config_file:
        output_config_file.write(_default_config_content)



from fanc.hic import LegacyHic, Hic
from fanc.matrix import Edge
from genomic_regions import GenomicRegion


def sample_hic(file_name=None, tmpdir=None):
    hic = Hic(file_name=file_name, tmpdir=tmpdir, mode='w')

    # add some nodes (12 to be exact)
    nodes = []
    for i in range(1, 5000, 1000):
        nodes.append(GenomicRegion(chromosome="chr1", start=i, end=i+1000-1))
    for i in range(1, 3000, 1000):
        nodes.append(GenomicRegion(chromosome="chr2", start=i, end=i+1000-1))
    for i in range(1, 2000, 500):
        nodes.append(GenomicRegion(chromosome="chr3", start=i, end=i+1000-1))
    hic.add_regions(nodes)

    # add some edges with increasing weight for testing
    edges = []
    weight = 1
    for i in range(0, len(nodes)):
        for j in range(i, len(nodes)):
            edges.append(Edge(source=i, sink=j, weight=weight))
            weight += 1

    hic.add_edges(edges)

    return hic


def sample_hic_big(file_name=None, tmpdir=None):
    hic = Hic(file_name=file_name, tmpdir=tmpdir, mode='w')

    # add some nodes (120 to be exact)
    nodes = []
    for i in range(1, 50000, 1000):
        nodes.append(GenomicRegion(chromosome="chr1", start=i, end=i + 1000 - 1))
    for i in range(1, 30000, 1000):
        nodes.append(GenomicRegion(chromosome="chr2", start=i, end=i + 1000 - 1))
    for i in range(1, 20000, 500):
        nodes.append(GenomicRegion(chromosome="chr3", start=i, end=i + 1000 - 1))
    hic.add_regions(nodes)

    # add some edges with increasing weight for testing
    edges = []
    for i in range(0, len(nodes)):
        for j in range(i, len(nodes)):
            weight = i + len(nodes) - (j - i)
            edges.append(Edge(source=i, sink=j, weight=weight))

    hic.add_edges(edges)

    return hic


def sample_fa_hic(file_name=None, zero_indices=set(), tmpdir=None):
    hic = Hic(file_name=file_name, tmpdir=tmpdir, mode='w')

    # add some nodes (120 to be exact)
    nodes = []
    for i in range(1, 50000, 1000):
        nodes.append(GenomicRegion(chromosome="chr1", start=i, end=i + 1000 - 1))
    for i in range(1, 30000, 1000):
        nodes.append(GenomicRegion(chromosome="chr2", start=i, end=i + 1000 - 1))
    for i in range(1, 20000, 500):
        nodes.append(GenomicRegion(chromosome="chr3", start=i, end=i + 1000 - 1))
    hic.add_regions(nodes)

    # add some edges with increasing weight for testing
    edges = []
    weight = 1
    for i in range(0, len(nodes)):
        for j in range(i, len(nodes)):
            if i not in zero_indices and j not in zero_indices:
                edges.append(Edge(source=i, sink=j, weight=weight))
            weight += 1

    hic.add_edges(edges)

    return hic


def sample_homogenous_hic(file_name=None, fill_value=1.0, zero_indices=set(), tmpdir=None):
    hic = Hic(file_name=file_name, tmpdir=tmpdir, mode='w')

    # add some nodes (120 to be exact)
    nodes = []
    for i in range(1, 50000, 1000):
        nodes.append(GenomicRegion(chromosome="chr1", start=i, end=i + 1000 - 1))
    for i in range(1, 30000, 1000):
        nodes.append(GenomicRegion(chromosome="chr2", start=i, end=i + 1000 - 1))
    for i in range(1, 20000, 500):
        nodes.append(GenomicRegion(chromosome="chr3", start=i, end=i + 1000 - 1))
    hic.add_regions(nodes)

    # add some edges with increasing weight for testing
    edges = []
    for i in range(0, len(nodes)):
        for j in range(i, len(nodes)):
            if i not in zero_indices and j not in zero_indices:
                edges.append(Edge(source=i, sink=j, weight=fill_value))

    hic.add_edges(edges)

    return hic


def sample_hic_matrix1(file_name=None, tmpdir=None):
    #     0 1 2 3 4 5 6 7 8 9
    #   #####################
    # 0 # 0 1 0 2 0 3 0 4 0 5
    # 1 #   6 0 7 0 8 0 9 0 1
    # 2 #     2 3 4 0 5 0 0 6
    # 3 #       7 0 8 9 0 1 0
    # 4 #         0 2 3 0 0 4
    # 5 #           5 6 7 8 9
    # 6 #             1 0 0 0
    # 7 #               2 3 0
    # 8 #                 0 4
    # 9 #                   5
    nodes = [
        GenomicRegion('chr1', 1, 1000),
        GenomicRegion('chr1', 1001, 2000),
        GenomicRegion('chr1', 2001, 3000),
        GenomicRegion('chr1', 3001, 4000),
        GenomicRegion('chr1', 4001, 5000),
        GenomicRegion('chr1', 5001, 6000),
        GenomicRegion('chr1', 6001, 7000),
        GenomicRegion('chr1', 7001, 8000),
        GenomicRegion('chr1', 8001, 9000),
        GenomicRegion('chr1', 9001, 10000)
    ]

    edges = [
        Edge(source=0, sink=1, weight=1), Edge(source=3, sink=5, weight=8),
        Edge(source=0, sink=3, weight=2), Edge(source=3, sink=6, weight=9),
        Edge(source=0, sink=5, weight=3), Edge(source=3, sink=8, weight=1),
        Edge(source=0, sink=7, weight=4), Edge(source=4, sink=5, weight=2),
        Edge(source=0, sink=9, weight=5), Edge(source=4, sink=6, weight=3),
        Edge(source=1, sink=1, weight=6), Edge(source=4, sink=9, weight=4),
        Edge(source=1, sink=3, weight=7), Edge(source=5, sink=5, weight=5),
        Edge(source=1, sink=5, weight=8), Edge(source=5, sink=6, weight=6),
        Edge(source=1, sink=7, weight=9), Edge(source=5, sink=7, weight=7),
        Edge(source=1, sink=9, weight=1), Edge(source=5, sink=8, weight=8),
        Edge(source=2, sink=2, weight=2), Edge(source=5, sink=9, weight=9),
        Edge(source=2, sink=3, weight=3), Edge(source=6, sink=6, weight=1),
        Edge(source=2, sink=4, weight=4), Edge(source=7, sink=7, weight=2),
        Edge(source=2, sink=6, weight=5), Edge(source=7, sink=8, weight=3),
        Edge(source=2, sink=9, weight=6), Edge(source=8, sink=9, weight=4),
        Edge(source=3, sink=3, weight=7), Edge(source=9, sink=9, weight=5)
    ]

    hic = Hic(file_name=file_name, tmpdir=tmpdir)
    hic.add_regions(nodes)
    hic.add_edges(edges)

    return hic


def sample_hic_matrix2(file_name=None, tmpdir=None):
    #     0 1 2 3 4 5 6 7 8 9
    #   #####################
    # 0 # 0 1 0 2 - 3 0 - 0 5
    # 1 #   6 0 7 - 8 0 - 0 1
    # 2 #     2 3 - 0 5 - 0 6
    # 3 #       7 - 8 9 - 1 0
    # 4 #         - - - - - -
    # 5 #           5 6 - 8 9
    # 6 #             1 - 0 0
    # 7 #               - - -
    # 8 #                 0 4
    # 9 #                   5
    nodes = [
        GenomicRegion('chr1', 1, 1000),
        GenomicRegion('chr1', 1001, 2000),
        GenomicRegion('chr1', 2001, 3000),
        GenomicRegion('chr1', 3001, 4000),
        GenomicRegion('chr1', 4001, 5000),
        GenomicRegion('chr1', 5001, 6000),
        GenomicRegion('chr1', 6001, 7000),
        GenomicRegion('chr1', 7001, 8000),
        GenomicRegion('chr1', 8001, 9000),
        GenomicRegion('chr1', 9001, 10000)
    ]

    edges = [
        Edge(source=0, sink=1, weight=1), Edge(source=3, sink=5, weight=8),
        Edge(source=0, sink=3, weight=2), Edge(source=3, sink=6, weight=9),
        Edge(source=0, sink=5, weight=3), Edge(source=3, sink=8, weight=1),
        Edge(source=0, sink=9, weight=5),
        Edge(source=1, sink=1, weight=6),
        Edge(source=1, sink=3, weight=7), Edge(source=5, sink=5, weight=5),
        Edge(source=1, sink=5, weight=8), Edge(source=5, sink=6, weight=6),
        Edge(source=1, sink=9, weight=1), Edge(source=5, sink=8, weight=8),
        Edge(source=2, sink=2, weight=2), Edge(source=5, sink=9, weight=9),
        Edge(source=2, sink=3, weight=3), Edge(source=6, sink=6, weight=1),
        Edge(source=2, sink=6, weight=5),
        Edge(source=2, sink=9, weight=6), Edge(source=8, sink=9, weight=4),
        Edge(source=3, sink=3, weight=7), Edge(source=9, sink=9, weight=5)
    ]

    hic = Hic(file_name=file_name, tmpdir=tmpdir)
    hic.add_regions(nodes)
    hic.add_edges(edges)

    return hic



from genomic_regions import GenomicRegion, Bedpe
from collections import defaultdict
import numpy as np
from ..tools.general import RareUpdateProgressBar
from ..general import FileGroup
from skimage.transform import resize
import warnings
import tables

import logging

logger = logging.getLogger(__name__)


class AggregateMatrix(FileGroup):
    """
    Construct and store aggregate matrices from matrix-based objects.

    Methods in this class can be used to generate various kinds of aggregate
    matrices, constructed from averaging the signal from different regions
    of a Hi-C (or similar) matrix. Particularly useful is the creation of
    aggregate matrices from observed/expected data.

    Class methods control how exactly an aggregate matrix is constructed:

    - :func:`AggregateMatrix.from_center` will aggregate Hi-C matrix regions along the diagonal
      in a fixed window around the region center. This is useful, for example,
      to observe the signal around TAD boundaries or other local features,
      such as the start of genes, enhancer locations, ...

    - :func:`AggregateMatrix.from_regions` will extract sub-matrices using
      regions of variable size - such as TADs - and interpolate them to the
      same number of pixels before aggregating them.

    - :func:`AggregateMatrix.from_center_pairs` will extract arbitrary Hi-C
      submatrices from a list of region pairs (representing row and column
      of the matrix). Each submatrix is centered on each region,
      and a fixed number of pixels around the center is extracted. This is
      used, for example, to plot aggregate matrices around loops, using the
      loop anchors as input.

    """

    _classid = 'AGGREGATEMATRIX'

    def __init__(self, file_name=None, mode='r', tmpdir=None, x=None, y=None):
        FileGroup.__init__(self, 'aggregate', file_name=file_name, mode=mode, tmpdir=tmpdir)

        self.x = x
        self.y = y

    def matrix(self, m=None):
        """
        Retrieve or set the aggregate matrix in this object.

        :param m: Numpy matrix
        :return: aggregate matrix
        """
        if m is not None:
            try:
                self.file.remove_node(self._group, 'aggregate_matrix')
            except tables.NoSuchNodeError:
                pass
            am = self.file.create_carray(self._group, 'aggregate_matrix', tables.Float32Atom(), m.shape)
            am[:] = m

        return self._group.aggregate_matrix[:]

    def region_pairs(self, pairs=None):
        """
        Retrieve or set the regions used to generate the aggregate matrix.

        :param pairs: Iterable of region tuples of the form
                      [(region1, region2), (region3, region4), ...].
                      If None, simply return the region pairs in this object.
        :return: List of region pairs [(region1, region2), (region3, region4), ...].
        """
        if pairs is not None:
            try:
                self.file.remove_node(self._group, 'region_pairs')
            except tables.NoSuchNodeError:
                pass

            pairs_table = self.file.create_table(self._group, 'region_pairs',
                                                 description={
                                                     'chromosome1': tables.StringCol(50, pos=0),
                                                     'start1': tables.Int32Col(pos=1),
                                                     'end1': tables.Int32Col(pos=2),
                                                     'strand1': tables.Int32Col(pos=3),
                                                     'chromosome2': tables.StringCol(50, pos=4),
                                                     'start2': tables.Int32Col(pos=5),
                                                     'end2': tables.Int32Col(pos=6),
                                                     'strand2': tables.Int32Col(pos=7)
                                                 })

            row = pairs_table.row
            for r1, r2 in pairs:
                row['chromosome1'] = r1.chromosome
                row['start1'] = r1.start
                row['end1'] = r1.end
                row['strand1'] = r1.strand if r1.strand is not None else 0

                row['chromosome2'] = r2.chromosome
                row['start2'] = r2.start
                row['end2'] = r2.end
                row['strand2'] = r2.strand if r2.strand is not None else 0

                row.append()

            pairs_table.flush()

        pairs = []
        pairs_table = self.file.get_node(self._group, 'region_pairs')
        for row in pairs_table.iterrows():
            r1 = GenomicRegion(chromosome=row['chromosome1'], start=row['start1'],
                               end=row['end1'], strand=row['strand1'])
            r2 = GenomicRegion(chromosome=row['chromosome2'], start=row['start2'],
                               end=row['end2'], strand=row['strand2'])
            pairs.append((r1, r2))
        return pairs

    def components(self, components=None):
        """
        Retrieve or store each individual submatrix composing the aggregate matrix.

        :param components: List of (masked) numpy arrays
        :return: List of (masked) numpy arrays
        """
        if components is not None:
            try:
                self.file.remove_node(self._group, 'components', recursive=True)
            except tables.NoSuchNodeError:
                pass

            component_group = self.file.create_group(self._group, 'components')

            for i, m in enumerate(components):
                if m is None:
                    m = np.array([np.nan])
                cm = self.file.create_carray(component_group, 'component_{}'.format(i),
                                             tables.Float32Atom(), m.shape)
                cm[:] = m

                if hasattr(m, 'mask'):
                    mm = self.file.create_carray(component_group, 'mask_{}'.format(i),
                                                 tables.BoolAtom(), m.shape)
                    mm[:] = m.mask
            self.file.flush()

        max_ix = -1
        masks = dict()
        components = dict()
        component_group = self.file.get_node(self._group, 'components')
        for node in self.file.iter_nodes(component_group):
            if node.name.startswith('mask_'):
                ix = int(node.name[5:])
                masks[ix] = node[:]
                max_ix = max(ix, max_ix)
            elif node.name.startswith('component_'):
                ix = int(node.name[10:])
                m = node[:]
                if m.shape == (1,) and np.isnan(m[0]):
                    components[ix] = None
                else:
                    components[ix] = m
                max_ix = max(ix, max_ix)

        sorted_components = []
        for ix in range(max_ix + 1):
            component = components[ix]
            if component is not None:
                if ix in masks:
                    mask = masks[ix]
                else:
                    mask = None

                sorted_components.append(np.ma.masked_array(component, mask=mask))
            else:
                sorted_components.append(None)
        return sorted_components

    @classmethod
    def from_center(cls, matrix, regions, window=200000,
                    rescale=False, scaling_exponent=-0.25,
                    keep_components=True,
                    file_name=None, tmpdir=None, region_viewpoint='center',
                    **kwargs):
        """
        Construct an aggregate matrix from square regions along
        the diagonal with a fixed window size.

        By default, the submatrix that is extracted from :code:`matrix` is
        centred on the region centre and has a window size specified by
        :code:`window`. You can change where the window will be centered using
        :code:`region_viewpoint`, which can be any of "center", "start", "end",
        "five_prime", or "three_prime". The latter two may be particularly useful
        for genomic features such as genes.

        Example for TAD boundaries:

        .. code::

            import fanc
            hic = fanc.load("/path/to/matrix.hic")
            tad_boundaries = fanc.load("/path/to/tad_boundaries.bed")

            # run aggregate analysis
            am = fanc.AggregateMatrix.from_center(hic, tad_boundaries.regions,
                                                  window=500000)
            # extract matrix when done
            m = am.matrix()

        :param matrix: An object of type :class:`RegionMatrixContainer`, such as a
                       Hic matrix
        :param regions: A list of :class:`GenomicRegion` objects
        :param window: A window size in base pairs
        :param rescale: If True, will use :code:`scaling_exponent` to artificially rescale
                        the aggregate matrix values using a power law
        :param scaling_exponent: The power law exponent used if :code:`rescale` is True
        :param keep_components: If True (default) will store each submatrix used
                                to generate the aggregate matrix in the
                                :class:`AggregateMatrix` object, which can be retrieved
                                using :func:`AggregateMatrix.components`
        :param file_name: If provided, stores the aggregate matrix object at this location.
        :param tmpdir: If True will work in temporary directory until the object is closed
        :param region_viewpoint: point on which window is centred. any of "center", "start", "end",
                                 "five_prime", or "three_prime"
        :param kwargs: Keyword arguments passed to :func:`extract_submatrices`
        :return: aggregate matrix
        """
        kwargs.setdefault('oe', True)
        kwargs.setdefault('keep_invalid', False)
        kwargs.setdefault('log', True)

        region_pairs = []
        for region in regions:
            new_start = int(getattr(region, region_viewpoint) - int(window / 2))
            new_end = int(getattr(region, region_viewpoint) + int(window / 2))
            new_region = GenomicRegion(chromosome=region.chromosome, start=new_start, end=new_end,
                                       strand=region.strand)
            region_pairs.append((new_region, new_region))

        component_regions = []
        component_matrices = []
        counter_matrix = None
        matrix_sum = None
        for (r1, r2), m in extract_submatrices(matrix, region_pairs, **kwargs):
            if counter_matrix is None:
                shape = m.shape
                counter_matrix = np.zeros(shape)
                matrix_sum = np.zeros(shape)

            if hasattr(m, 'mask'):
                inverted_mask = ~m.mask
                counter_matrix += inverted_mask.astype('int')
            else:
                counter_matrix += np.ones(counter_matrix.shape)

            matrix_sum += m

            component_regions.append((r1, r2))
            if keep_components:
                component_matrices.append(m)

        am = matrix_sum / counter_matrix

        if rescale:
            am = _rescale_oe_matrix(am, matrix.bin_size, scaling_exponent=scaling_exponent)

        aggregate_object = cls(file_name=file_name, mode='w', tmpdir=tmpdir)
        aggregate_object.matrix(am)
        aggregate_object.region_pairs(component_regions)
        aggregate_object.components(component_matrices)

        return aggregate_object

    @classmethod
    def from_regions(cls, hic, tad_regions, pixels=90,
                     rescale=False, scaling_exponent=-0.25,
                     interpolation=0,
                     boundary_mode='reflect', keep_mask=True,
                     absolute_extension=0, relative_extension=1.0,
                     keep_components=True, anti_aliasing=True,
                     file_name=None, tmpdir=None,
                     **kwargs):
        """
        Construct aggregate matrix from variable regions along the diagonal.

        For each region in :code:`tad_regions`, a submatrix is extracted and
        interpolated so that it is exactly :code:`pixels` x :code:`pixels` big.
        You can expand each region by a relative amount using :code:`relative_extension`.

        Example for aggregate TADs:

        .. code::

            import fanc
            hic = fanc.load("/path/to/matrix.hic")
            tads = fanc.load("/path/to/tads.bed")

            # run aggregate analysis
            am = fanc.AggregateMatrix.from_regions(hic, tads.regions,
                                                   relative_extension=3.)
            # extract matrix when done
            m = am.matrix()  # 90x90 matrix with aggregate TAD in the centre


        :param hic: An object of type :class:`RegionMatrixContainer`, such as a
                       Hic matrix
        :param tad_regions: A list of :class:`GenomicRegion` objects
        :param pixels: Number of pixels along each dimension of the aggregate matrix
        :param rescale: If True, will use :code:`scaling_exponent` to artificially rescale
                        the aggregate matrix values using a power law
        :param scaling_exponent: The power law exponent used if :code:`rescale` is True
        :param interpolation: Type of interpolation used on each submatrix in range 0-5.
                              0: Nearest-neighbor (default), 1: Bi-linear, 2: Bi-quadratic,
                              3: Bi-cubic, 4: Bi-quartic, 5: Bi-quintic
        :param boundary_mode: Points outside the boundaries of the input are filled
                              according to the given mode. Options are constant, edge,
                              symmetrix, reflect, and warp. Affects submatrix interpolation.
        :param keep_mask: If True (default) maksed Hi-C regions will also be interpolated.
        :param absolute_extension: Absolute number of base pairs by which to expand each
                                   region
        :param absolute_extension: Amount by which to expand each region as a fraction
                                   of each region. Values smaller than 1 lead to region
                                   shrinking
        :param keep_components: If True (default) will store each submatrix used
                                to generate the aggregate matrix in the
                                :class:`AggregateMatrix` object, which can be retrieved
                                using :func:`AggregateMatrix.components`
        :param file_name: If provided, stores the aggregate matrix object at this location.
        :param tmpdir: If True will work in temporary directory until the object is closed
        :param kwargs: Keyword argumnts passed to :func:`extract_submatrices`

        :return: aggregate matrix
        """
        kwargs.setdefault('oe', True)
        kwargs.setdefault('keep_invalid', False)
        kwargs.setdefault('log', True)

        component_regions = []
        component_matrices = []
        shape = (pixels, pixels)
        counter_matrix = np.zeros(shape)
        matrix_sum = np.zeros(shape)
        for (r1, r2), m in _tad_matrix_iterator(hic, tad_regions,
                                                absolute_extension=absolute_extension,
                                                relative_extension=relative_extension,
                                                **kwargs):
            if m is not None:
                ms = resize(m, shape, mode=boundary_mode, anti_aliasing=anti_aliasing,
                            preserve_range=False, clip=False, order=interpolation)

                if keep_mask and hasattr(ms, 'mask'):
                    mask = resize(m.mask, shape, mode=boundary_mode, anti_aliasing=anti_aliasing,
                                  preserve_range=False, clip=False, order=interpolation).astype('bool')
                    ms = np.ma.masked_where(mask, ms)
                    inverted_mask = ~mask
                    counter_matrix += inverted_mask.astype('int')
                else:
                    counter_matrix += np.ones(shape)

                matrix_sum += ms

            component_regions.append((r1, r2))
            if keep_components:
                component_matrices.append(m)

        am = matrix_sum/counter_matrix

        if rescale:
            am = _rescale_oe_matrix(am, hic.bin_size, scaling_exponent=scaling_exponent)

        aggregate_object = cls(file_name=file_name, mode='w', tmpdir=tmpdir)
        aggregate_object.matrix(am)
        aggregate_object.region_pairs(component_regions)
        aggregate_object.components(component_matrices)

        return aggregate_object

    @classmethod
    def from_center_pairs(cls, hic, pair_regions, window=None, pixels=16,
                          keep_components=True, file_name=None, tmpdir=None,
                          region_viewpoint='center',
                          **kwargs):
        """

        :param hic:
        :param pair_regions:
        :param window:
        :param pixels:
        :param keep_components:
        :param file_name:
        :param tmpdir:
        :param region_viewpoint:
        :param kwargs:
        :return:
        """
        kwargs.setdefault('oe', True)
        kwargs.setdefault('keep_invalid', False)
        kwargs.setdefault('log', True)

        if window is not None:
            bin_size = hic.bin_size
            pixels = np.round(window/bin_size)

        component_regions = []
        component_matrices = []
        shape = (pixels, pixels)
        counter_matrix = np.zeros(shape)
        matrix_sum = np.zeros(shape)
        for (r1, r2), m in _loop_matrix_iterator(hic, pair_regions, pixels=pixels,
                                                 region_viewpoint=region_viewpoint,
                                                 **kwargs):
            try:
                if hasattr(m, 'mask'):
                    inverted_mask = ~m.mask
                    counter_matrix += inverted_mask.astype('int')
                else:
                    counter_matrix += np.ones(shape)
                matrix_sum += m
            except ValueError:
                warnings.warn("Regions {} vs {} did not produce a valid matrix for aggregation!"
                              .format(r1, r2))
                continue

            component_regions.append((r1, r2))
            if keep_components:
                component_matrices.append(m)

        am = matrix_sum/counter_matrix

        aggregate_object = cls(file_name=file_name, mode='w', tmpdir=tmpdir)
        aggregate_object.matrix(am)
        aggregate_object.region_pairs(component_regions)
        aggregate_object.components(component_matrices)

        return aggregate_object


def _aggregate_region_bins(hic, region, offset=0):
    region_slice = hic.region_bins(region)
    return region_slice.start-offset, region_slice.stop-offset


def extract_submatrices(matrix, region_pairs, oe=False,
                        log=True, cache=True, mask_inf=True,
                        keep_invalid=False, orient_strand=False,
                        **kwargs):
    cl = matrix.chromosome_lengths
    cb = matrix.chromosome_bins

    valid_region_pairs = defaultdict(list)
    invalid_region_pairs = list()
    logger.info("Checking region pair validity...")
    valid, invalid = 0, 0
    for ix, (region1, region2) in enumerate(region_pairs):
        is_invalid = False
        if region1 is None or region2 is None:
            is_invalid = True
        elif region1.start < 1 or region1.chromosome not in cl or region1.end > cl[region1.chromosome]:
            is_invalid = True
        elif region2.start < 1 or region2.chromosome not in cl or region2.end > cl[region2.chromosome]:
            is_invalid = True
        if is_invalid:
            invalid += 1
            invalid_region_pairs.append((ix, region1, region2))
        else:
            valid += 1
            valid_region_pairs[(region1.chromosome, region2.chromosome)].append((ix, region1, region2))
    logger.info("{}/{} region pairs are invalid".format(invalid, valid + invalid))

    intra_expected, inter_expected = dict(), None
    if oe:
        logger.debug("Calculating expected values...")
        _, intra_expected, inter_expected = matrix.expected_values()

    order = []
    matrices = []
    final_regions = []
    with RareUpdateProgressBar(max_value=valid, prefix='Matrices') as pb:
        current_matrix = 0
        for (chromosome1, chromosome2), regions_pairs_by_chromosome in valid_region_pairs.items():
            if cache:
                sub_matrix = matrix.matrix((chromosome1, chromosome2), **kwargs)
                offset1 = cb[chromosome1][0]
                offset2 = cb[chromosome2][0]
            else:
                sub_matrix = matrix
                offset1 = 0
                offset2 = 0

            for (region_ix, region1, region2) in regions_pairs_by_chromosome:
                final_regions.append((region1, region2))
                current_matrix += 1
                region1_bins = _aggregate_region_bins(matrix, region1, offset1)
                region2_bins = _aggregate_region_bins(matrix, region2, offset2)

                if cache:
                    ms = sub_matrix[region1_bins[0]:region1_bins[1], region2_bins[0]: region2_bins[1]]
                    m = ms.copy()
                    del ms
                else:
                    s1 = slice(region1_bins[0], region1_bins[1])
                    s2 = slice(region2_bins[0], region2_bins[1])
                    m = matrix.matrix((s1, s2), **kwargs)

                if oe:
                    e = np.ones(m.shape)
                    if chromosome1 != chromosome2:
                        e.fill(inter_expected)
                    else:
                        for i, row in enumerate(range(region1_bins[0], region1_bins[1])):
                            for j, col in enumerate(range(region2_bins[0], region2_bins[1])):
                                ix = abs(col - row)
                                e[i, j] = intra_expected[chromosome1][ix]

                    if log:
                        m = np.log2(m/e)
                        m[np.isnan(m)] = 0.
                    else:
                        m = m/e
                        m[np.isnan(m)] = 1

                if mask_inf:
                    m_mask = np.isinf(m)
                    if not hasattr(m, 'mask'):
                        m = np.ma.masked_where(m_mask, m)
                    m.mask += m_mask

                if orient_strand and region1.is_reverse() and region2.is_reverse():
                    m = np.flip(np.flip(m, 0), 1)

                pb.update(current_matrix)
                matrices.append(m)
                order.append(region_ix)

            if cache:
                del sub_matrix

    if keep_invalid:
        for region_ix, r1, r2 in invalid_region_pairs:
            matrices.append(None)
            order.append(region_ix)
            final_regions.append((r1, r2))

    final_regions = [final_regions[ix] for ix in np.argsort(order)]
    matrices = [matrices[ix] for ix in np.argsort(order)]

    return zip(final_regions, matrices)


def _rescale_oe_matrix(matrix, bin_size, scaling_exponent=-0.25):
    rm = np.zeros(matrix.shape)
    b = bin_size
    for i in range(matrix.shape[0]):
        for j in range(matrix.shape[1]):
            v = (abs(i - j) * b + b) ** scaling_exponent
            try:
                if matrix.mask[i, j]:
                    continue
            except AttributeError:
                pass
            rm[i, j] = v * matrix[i, j]
            rm[j, i] = v * matrix[j, i]
    return rm


def aggregate_boundaries(hic, boundary_regions, window=200000,
                         rescale=False, scaling_exponent=-0.25,
                         **kwargs):
    kwargs.setdefault('norm', True)
    kwargs.setdefault('keep_invalid', False)
    kwargs.setdefault('log', True)

    region_pairs = []
    for region in boundary_regions:
        new_start = int(region.center - int(window / 2))
        new_end = int(region.center + int(window / 2))
        new_region = GenomicRegion(chromosome=region.chromosome, start=new_start, end=new_end,
                                   strand=region.strand)
        region_pairs.append((new_region, new_region))

    counter_matrix = None
    matrix_sum = None
    for m in extract_submatrices(hic, region_pairs, **kwargs):
        if counter_matrix is None:
            shape = m.shape
            counter_matrix = np.zeros(shape)
            matrix_sum = np.zeros(shape)

        if hasattr(m, 'mask'):
            inverted_mask = ~m.mask
            counter_matrix += inverted_mask.astype('int')
        else:
            counter_matrix += np.ones(counter_matrix.shape)

        matrix_sum += m

    am = matrix_sum / counter_matrix

    if rescale:
        am = _rescale_oe_matrix(am, hic.bin_size, scaling_exponent=scaling_exponent)

    return am


def _tad_matrix_iterator(hic, tad_regions, absolute_extension=0, relative_extension=1., **kwargs):
    region_pairs = []
    for region in tad_regions:
        new_region = region.expand(absolute=absolute_extension, relative=relative_extension)
        region_pairs.append((new_region, new_region))

    for pair, m in extract_submatrices(hic, region_pairs, **kwargs):
        yield pair, m


def aggregate_tads(hic, tad_regions, pixels=90, rescale=False, scaling_exponent=-0.25,
                   boundary_mode='reflect', keep_mask=True,
                   absolute_extension=0, relative_extension=1.0,
                   anti_aliasing=True,
                   **kwargs):
    kwargs.setdefault('norm', True)
    kwargs.setdefault('keep_invalid', False)
    kwargs.setdefault('log', True)

    shape = (pixels, pixels)
    counter_matrix = np.zeros(shape)
    matrix_sum = np.zeros(shape)
    for pair, m in _tad_matrix_iterator(hic, tad_regions,
                                        absolute_extension=absolute_extension,
                                        relative_extension=relative_extension, **kwargs):

        ms = resize(m, shape, mode=boundary_mode, anti_aliasing=anti_aliasing,
                    preserve_range=False, clip=False)
        if ms.shape != shape:
            raise ValueError("Matrix shape not as expected ({}) vs ({}). "
                             "This may be due to the recent replacement of 'imresize' with 'zoom'. "
                             "Please contact Kai.")

        if keep_mask and hasattr(ms, 'mask'):
            # mask = imresize(m.mask, shape, interp='nearest').astype('bool')
            mask = resize(m.mask, shape, mode=boundary_mode, anti_aliasing=anti_aliasing,
                          preserve_range=False, clip=False).astype('bool')
            ms = np.ma.masked_where(mask, ms)
            inverted_mask = ~mask
            counter_matrix += inverted_mask.astype('int')
        else:
            counter_matrix += np.ones(shape)

        matrix_sum += ms

    am = matrix_sum/counter_matrix

    if rescale:
        am = _rescale_oe_matrix(am, hic.bin_size, scaling_exponent=scaling_exponent)

    return am


def tad_strength(hic, tad_regions=None, **kwargs):
    kwargs.setdefault('norm', True)
    kwargs.setdefault('keep_invalid', False)
    kwargs.setdefault('log', False)
    kwargs['relative_extension'] = 1.
    kwargs['absolute_extension'] = 0
    kwargs['keep_components'] = True
    kwargs['oe'] = True

    if isinstance(hic, AggregateMatrix):
        aggregate_tads = hic
    else:
        if tad_regions is None:
            raise ValueError("tad_regions cannot be None when providing matrix as input!")
        aggregate_tads = AggregateMatrix.from_regions(hic, tad_regions, **kwargs)

    tad_strengths = []
    for m in aggregate_tads.components():
        if m is None:
            if kwargs['keep_invalid']:
                tad_strengths.append(np.nan)
            continue

        tl = int(m.shape[0]/3)
        upper_third = slice(0, tl)
        middle_third = slice(tl, 2*tl)
        lower_third = slice(2*tl, m.shape[0])
        tad_sum = np.nansum(m[middle_third, middle_third]) / np.nansum(np.logical_not(m.mask)[middle_third, middle_third])
        upper_sum = np.nansum(m[upper_third, middle_third]) / np.nansum(np.logical_not(m.mask)[upper_third, middle_third])
        lower_sum = np.nansum(m[lower_third, upper_third]) / np.nansum(np.logical_not(m.mask)[lower_third, upper_third])
        try:
            ts = float(tad_sum / ((upper_sum + lower_sum) / 2))
        except ZeroDivisionError:
            if kwargs['keep_invalid']:
                tad_strengths.append(np.nan)
            continue
        tad_strengths.append(np.log2(ts))
    return tad_strengths


def _loop_regions_from_bedpe(bedpe):
    anchors = []
    for region in bedpe.regions:
        a1 = GenomicRegion(chromosome=region.chromosome1, start=region.start1, end=region.end1)
        a2 = GenomicRegion(chromosome=region.chromosome2, start=region.start2, end=region.end2)
        anchors.append((a1, a2))
    return anchors


def _loop_matrix_iterator(hic, loop_regions, pixels=16,
                          region_viewpoint='center', **kwargs):
    left = int(pixels / 2)
    right = left if pixels % 2 == 1 else left - 1

    if isinstance(loop_regions, Bedpe):
        loop_regions = _loop_regions_from_bedpe(loop_regions)

    bin_size = hic.bin_size
    region_pairs = []
    invalid = 0
    for (anchor1, anchor2) in loop_regions:
        a1 = GenomicRegion(chromosome=anchor1.chromosome,
                           start=getattr(anchor1, region_viewpoint),
                           end=getattr(anchor1, region_viewpoint))
        a2 = GenomicRegion(chromosome=anchor2.chromosome,
                           start=getattr(anchor2, region_viewpoint),
                           end=getattr(anchor2, region_viewpoint))

        try:
            r1 = list(hic.regions(a1))[0].copy()
            r2 = list(hic.regions(a2))[0].copy()
            r1.start -= left * bin_size
            r1.end += right * bin_size
            r2.start -= left * bin_size
            r2.end += right * bin_size
            region_pairs.append((r1, r2))
        except IndexError:
            invalid += 1
            region_pairs.append((None, None))

    if invalid > 0:
        logger.warning("{} region pairs invalid, most likely due to missing chromosome data".format(invalid))

    for pair, m in extract_submatrices(hic, region_pairs, **kwargs):
        yield pair, m


def aggregate_loops(hic, loop_regions, pixels=16, **kwargs):
    kwargs.setdefault('norm', True)
    kwargs.setdefault('keep_invalid', False)
    kwargs.setdefault('log', True)

    shape = (pixels, pixels)
    counter_matrix = np.zeros(shape)
    matrix_sum = np.zeros(shape)
    for m in _loop_matrix_iterator(hic, loop_regions, pixels=pixels, **kwargs):
        if hasattr(m, 'mask'):
            inverted_mask = ~m.mask
            counter_matrix += inverted_mask.astype('int')
        else:
            counter_matrix += np.ones(shape)
        matrix_sum += m

    return matrix_sum/counter_matrix


def loop_strength(hic, loop_regions, pixels=16, **kwargs):
    kwargs.setdefault('log', False)
    kwargs.setdefault('norm', True)
    kwargs.setdefault('oe', True)

    try:
        include_nan = kwargs.pop('keep_invalid')
    except KeyError:
        include_nan = False

    if isinstance(loop_regions, Bedpe):
        loop_regions = _loop_regions_from_bedpe(loop_regions)

    # generating new regions
    new_region_pairs = []  # 0: original, 1: control left, 2: control right
    for (region1, region2) in loop_regions:
        d = int(abs(region1.center - region2.center))
        new_left = GenomicRegion(chromosome=region1.chromosome, start=region1.start - d, end=region1.end - d)
        new_right = GenomicRegion(chromosome=region1.chromosome, start=region2.start + d, end=region2.end + d)
        new_region_pairs.append((region1, region2))
        new_region_pairs.append((new_left, region1))
        new_region_pairs.append((region2, new_right))

    original, left, right = [], [], []
    for i, (pair, m) in enumerate(_loop_matrix_iterator(hic, new_region_pairs, pixels=pixels,
                                                        keep_invalid=True, **kwargs)):
        if m is not None:
            value = float(np.nansum(m)/np.nansum(np.logical_not(m.mask)))
        else:
            value = None

        if i % 3 == 0:
            original.append(value)
        elif i % 3 == 1:
            left.append(value)
        else:
            right.append(value)

    ratios = []
    for i in range(len(original)):
        if original[i] is None or (left[i] is None and right[i] is None):
            if include_nan:
                ratios.append(np.nan)
            continue

        try:
            if left[i] is None:
                r = original[i]/right[i]
            elif right[i] is None:
                r = original[i]/left[i]
            else:
                r = original[i]/((left[i]+right[i])/2)
            ratios.append(np.log2(r))
        except ZeroDivisionError:
            if include_nan:
                ratios.append(np.nan)
    return ratios


def contact_directionality_bias(hic, regions, distance=1000000, region_anchor='center', **kwargs):
    forward_region_pairs = []
    reverse_region_pairs = []
    for region in regions:
        pos = int(getattr(region, region_anchor))
        new_region = GenomicRegion(chromosome=region.chromosome, start=pos, end=pos, strand=region.strand)
        if region.is_forward():
            forward_region_pairs.append((new_region, new_region.expand(absolute=distance)))
        else:
            reverse_region_pairs.append((new_region, new_region.expand(absolute=distance)))

    cumulative_forward = np.zeros(int(distance / hic.bin_size) * 2 + 1)
    count_forward = np.zeros(int(distance / hic.bin_size) * 2 + 1)
    for matrix in extract_submatrices(hic, forward_region_pairs, **kwargs):
        cumulative_forward += matrix[0, :]
        if hasattr(matrix, 'mask'):
            inverted_mask = ~matrix.mask
            count_forward += inverted_mask.astype('int')[0, :]
        else:
            count_forward += np.ones(count_forward.shape)

    cumulative_reverse = np.zeros(int(distance / hic.bin_size) * 2 + 1)
    count_reverse = np.zeros(int(distance / hic.bin_size) * 2 + 1)
    for matrix in extract_submatrices(hic, reverse_region_pairs, **kwargs):
        cumulative_reverse += matrix[0, :]
        if hasattr(matrix, 'mask'):
            inverted_mask = ~matrix.mask
            count_reverse += inverted_mask.astype('int')[0, :]
        else:
            count_reverse += np.ones(count_reverse.shape)

    avg_forward = cumulative_forward / count_forward
    avg_reverse = cumulative_reverse / count_reverse

    bin_size = hic.bin_size
    d = []
    ratio_forward = []
    ratio_reverse = []
    center = int(len(avg_forward)/2)
    for ix in range(center + 1):
        d.append(ix * bin_size)
        ratio_forward.append(avg_forward[center + ix] / avg_forward[center - ix])
        ratio_reverse.append(avg_reverse[center + ix] / avg_reverse[center - ix])

    return d, ratio_forward, ratio_reverse



"""
Provide plotting functions for genomic data types.

Many common data types used in genomics research are supported. Including, but
not limited to, :class:`Hi-C <fanc.plotting.HicPlot>`,
:class:`bed <fanc.plotting.GenomicFeaturePlot>`,
:class:`bigwig <fanc.plotting.BigWigPlot>`
and :class:`gene (GTF) <fanc.plotting.GenePlot>` file visualization.
The basic idea is that figures can be composed of multiple panels which are
arranged vertically and share a common x-axis representing genomic coordinates.

Each panel is created separately and then combined into a single figure.
For example, when analyzing Hi-C data it is often interesting to correlate
features in the Hi-C map with ChIP-seq tracks. In that case, one would first
create a :class:`~fanc.plotting.HicPlot` object, which visualizes
Hi-C data, and then a :class:`~fanc.plotting.BigWigPlot` object, which
can plot bigwig files that are used during ChIP-seq analysis. Finally, the two
objects are used to create a :class:`~fanc.plotting.GenomicFigure`.

Example
-------

FAN-C comes with a few example datasets that can be used to explore the basic
functionalities of the plotting module. The paths in this example are relative
to the top-level fanc directory where the setup.py file is located.

.. note::
    The paths to the example datasets can be accessed easily using the
    example_data dictionary:

    .. code:: python

        import fanc
        print(fanc.example_data)

    Yields a dict with the user-specific path to the datasets:

    ::

        {'chip_bedgraph': '/Users/example/fanc/fanc/test/data/test_plotting/CTCF_ChIP_FE_chr11_77-80Mb_mouse_embryo_fibroblasts.bedgraph.gz',
         'chip_bigwig': '/Users/example/fanc/fanc/test/data/test_plotting/CTCF_ChIP_FE_chr11_77-80Mb_mouse_embryo_fibroblasts.bigwig',
         'chip_peak_bed': '/Users/example/fanc/fanc/test/data/test_plotting/CTCF_ChIP_FE_chr11_77-80Mb_mouse_embryo_fibroblasts.peaks.bed.gz',
         'gene_gtf': '/Users/example/fanc/fanc/test/data/test_plotting/genes_mm10_chr11_77-80Mb.gtf.gz',
         'hic': '/Users/example/fanc/fanc/test/data/test_network/rao2014.chr11_77400000_78600000.hic'}

.. code:: python

    import fanc.plotting as kplot

    # Create Hic plot
    hplot = kplot.HicPlot("fanc/test/data/test_network/rao2014.chr11_77400000_78600000.hic")

    # Create plot showing some CTCF ChIP-seq data from ENCODE
    bplot = kplot.BigWigPlot("fanc/test/data/test_plotting/CTCF_ChIP_FE_chr11_77-80Mb_mouse_embryo_fibroblasts.bigwig",
                             title="CTCF ChIP", ylabel="fold enrichment")

    # Create plot of all genes in the region. The squash option in combination
    # with the group_by option causes all exons of each gene to be merged.
    # This is useful if the number of alternative transcripts is overwhelming
    # the plot.
    gplot = kplot.GenePlot("fanc/test/data/test_plotting/genes_mm10_chr11_77-80Mb.gtf.gz",
                           group_by="gene_name", squash=True, show_labels=False,
                           title="mm10 genes")

    # The created plots are used to generate a figure by passing them as a list
    # to the GenomicFigure constructor. The order in which they are passed
    # determines the order of panels in the figure.
    gfig = kplot.GenomicFigure([hplot, bplot, gplot])

    # Plot a specific region of the genome
    fig, axes = gfig.plot("chr11:77400000-78600000")

    # Open plot in an interactive window
    fig.show()

    Example rendering of the above code saved using
    ``fig.savefig("example.png", dpi=100)``.

Editing figure and axes
-----------------------

The :meth:`GenomicFigure.plot() <fanc.plotting.GenomicFigure.plot>`
function returns standard matplotlib Figure and a list of Axes instances that
can be further adjusted using standard matplotlib methods. The matplotlib axes
instance associated with each plot is also accesible from the "ax" property of
each plot.

.. warning:: The Axes instances of the plot should only be edited after
    :meth:`GenomicFigure.plot() <fanc.plotting.GenomicFigure.plot>`
    has been called. Otherwise any changes that were made may be overwritten
    when the plot() method is called.

For example, to add a bit of annotating text at a specific location in the
BigWigPlot, the example above can be edited as follows:

.. code:: python

    fig, axes = gfig.plot("chr11:77400000-78600000")
    bplot.ax.text(77850000, 60, "Interesting peak!")
    fig.show()

The coordinates in the Axes are data coordinates, the x-axis is genomic
coordinates on the current chromosome and the y-axis in this case the
fold-enrichment of the bigwig track.

Basic Plot types and options
----------------------------

.. note::
    An explanation of each plot class and the parameters that it supports can be
    accessed by suffixing a question mark (in Ipython/Jupyter) or calling the
    help() function:

    .. code::

        import fanc.plotting as kplot
        kplot.BigWigPlot? # Ipython/Jupyter
        help(kplot.BigWigPlot) # standard python


A few basic parameters such as a title and the aspect ratio are available for
all plot classes. The aspect ratio parameter is a floating point number between
0 and 1 that determines the height of the plot. A value of 1 results in a square
plot, .5 represents a plot that is half as high as it is wide.

The :class:`~fanc.plotting.GenomicFigure` provides a few convenience
parameters. Setting ``ticks_last=True`` for example removes tick labels from all
panels but the last one which makes the overall plot more compact.

Independent x-axis and inverting x-axis
'''''''''''''''''''''''''''''''''''''''

By default, the x-axis of all plots in a figure are linked, meaning that all
plots display exactly the same region. In some situations in can be helpful
to plot a multiple regions, such as when features in syntenic regions across
multiple species need to be compared. Since syntenic regions can be on the +
strand in some species and on the - strande, sometimes the x-axis also needs to
be inverted to maintain correct orientation.

In this situation the ``independent-x`` option should be set in the
:class:`~fanc.plotting.GenomicFigure`. As a result, the
:meth:`GenomicFigure.plot() <fanc.plotting.GenomicFigure.plot>` method
no longer expects a single region as argument, but a list of regions equal to
the number of plots in the figure. We can modify the example above to illustrate
this point:

.. code:: python

    gfig = kplot.GenomicFigure([hplot, bplot, gplot], independent_x=True)
    fig, axes = gfig.plot(["chr11:77400000-78600000", "chr11:77500000-78600000",
                           "chr11:77200000-78600000"])
    fig.show()

Synchronize y-axis limits for multiple datasets
'''''''''''''''''''''''''''''''''''''''''''''''

Sometimes it can be useful to synchronize y-axis limits across multiple datasets
in order to compare their signals. Many plot types support plotting multiple
datasets in the same panel, making comparisons easy:

.. code:: python

    bplot = kplot.BigWigPlot(["dataset1.bigwig", "dataset2.bigwig"])

Alternatively it is possible to synchronize y-axis limits across panels using
:class:`~fanc.plotting.LimitGroup` instances. In this example the y-axis limits
are shared between bplot1 and bplot2 and seperately between bplot3 and bplot4:

.. code:: python

    ygroup = kplot.LimitGroup()
    ygroup2 = kplot.LimitGroup(limit=(None, 10))
    bplot1 = kplot.BigWigPlot("dataset1.bigwig", ylim=ygroup)
    bplot2 = kplot.BigWigPlot("dataset2.bigwig", ylim=ygroup)
    bplot3 = kplot.BigWigPlot("dataset3.bigwig", ylim=ygroup2)
    bplot4 = kplot.BigWigPlot("dataset4.bigwig", ylim=ygroup2)

It is also possible to constrain the y-axis limits. Passing ``limit=(None, 10)``
to the constructor constrains the upper limit to a maximum of 10 while leaving
the lower limit unconstrained.

Programmatic plotting using loops
---------------------------------

fanc plotting is ideally suited for programmatic generation of many plots or
dynamic assembly of multiple datasets in a single figure. In this example three
Hi-C datasets are visualized in a single figure:

.. code:: python

    import fanc.plotting as kplot
    hic_datasets = ["my_data1.hic", "my_data2.hic", my_data3.hic"]
    hic_plots = [kplot.HicPlot(h, max_dist=500000) for h in hic_datasets]
    gfig = kplot.GenomicFigure(hic_plots)
    regions = ["chr11:77400000-78600000", "chr11:1100000-13600000"]
    for r in regions:
        fig, axes = gfig.plot(r)
        fig.savefig("plot_region_{}.png".format(r.replace(":", "_")))

"""

from fanc.plotting.hic_plotter import HicPlot, HicPlot2D, HicComparisonPlot2D, \
    HicSlicePlot, HicPeakPlot, TriangularMatrixPlot, SquareMatrixPlot, SplitMatrixPlot
from fanc.plotting.plotter import VerticalSplitPlot, GenomicVectorArrayPlot, GenomicFeaturePlot, GenomicRegionsPlot, \
    GenomicFeatureScorePlot, GenomicFigure, BigWigPlot, GenePlot, LinePlot, \
    FeatureLayerPlot, GenomicDataFramePlot, HighlightAnnotation, RegionsValuesPlot, \
    BarPlot, Virtual4CPlot, MirrorMatrixPlot
from fanc.plotting.helpers import append_axes, absolute_wspace_hspace, SymmetricNorm, \
                                  style_ticks_whitegrid, LimitGroup

from fanc.plotting.colormaps import *
import seaborn as sns
import matplotlib.pyplot as plt

sns.set_style("ticks")

plt.register_cmap(name='viridis', cmap=viridis)
plt.register_cmap(name='plasma', cmap=plasma)
plt.register_cmap(name='inferno', cmap=inferno)
plt.register_cmap(name='magma', cmap=magma)
plt.register_cmap(name='RdBuWhitespace_r', cmap=fc_cmap)
plt.register_cmap(name='germany', cmap=germany_cmap)
plt.register_cmap(name='germany_r', cmap=germany_r)
plt.register_cmap(name='white_red', cmap=white_red)
plt.register_cmap(name='white_red_r', cmap=white_red_r)



import fanc
from fanc.config import config
from fanc.plotting.base_plotter import BasePlotterMatrix, BasePlotter1D, BasePlotter2D, ScalarDataPlot, \
                                       PlotMeta, BaseOverlayPlotter
from fanc.plotting.helpers import append_axes, style_ticks_whitegrid
from fanc.tools.general import str_to_int
from genomic_regions import GenomicRegion, as_region
from ..matrix import RegionMatrixTable, RegionMatrixContainer
from ..architecture.comparisons import SplitMatrix
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, CheckButtons
import matplotlib.gridspec as grd
import matplotlib.patches as patches
from abc import ABCMeta
import numpy as np
from scipy.ndimage.filters import gaussian_filter
import itertools as it
import types
import seaborn as sns
from future.utils import with_metaclass, string_types
from collections import defaultdict
from itertools import cycle
from ..matrix import RegionMatrix
from ..peaks import ObservedPeakFilter, FdrPeakFilter, EnrichmentPeakFilter, MappabilityPeakFilter
import logging
logger = logging.getLogger(__name__)


def prepare_hic_buffer(hic_data, buffering_strategy="relative", buffering_arg=1,
                       weight_field=None, default_value=None, smooth_sigma=None,
                       norm=True, oe=False, log=False):
    """
    Prepare :class:`~BufferedMatrix` from hic data.

    :param hic_data: :class:`~fanc.data.genomic.RegionMatrixTable` or
                     :class:`~fanc.data.genomic.RegionMatrix`
    :param buffering_strategy: "all", "fixed" or "relative"
                               "all" buffers the whole matrix
                               "fixed" buffers a fixed area, specified by buffering_arg
                                       around the query area
                               "relative" buffers a multiple of the query area.
                                          With buffering_arg=1 the query area plus
                                          the same amount upstream and downstream
                                          are buffered
    :param buffering_arg: Number specifying how much around the query area is buffered
    """
    if isinstance(hic_data, RegionMatrixContainer):
        return BufferedMatrix(hic_data, buffering_strategy=buffering_strategy,
                              buffering_arg=buffering_arg, weight_field=weight_field,
                              default_value=default_value, smooth_sigma=smooth_sigma,
                              norm=norm, oe=oe, log=log)
    else:
        raise ValueError("Unknown type for hic_data")


class BufferedMatrix(object):
    """
    Buffer contents of any :class:`~fanc.Hic` like objects. Matrix is
    prefetched and stored in memory. Buffer contents can quickly be fetched
    from memory. Different buffering strategies allow buffering of nearby
    regions so that adjacent parts of the matrix can quickly be fetched.
    """
    _STRATEGY_ALL = "all"
    _STRATEGY_FIXED = "fixed"
    _STRATEGY_RELATIVE = "relative"

    def __init__(self, data, buffering_strategy="relative", buffering_arg=1,
                 weight_field=None, default_value=None, smooth_sigma=None,
                 norm=True, oe=False, log=False):
        """
        Initialize a buffer for Matrix-like objects that support
        indexing using class:`~GenomicRegion` objects, such as class:`~fanc.Hic`
        or class:`~fanc.RegionMatrix` objects.

        :param data: Data to be buffered
        :param buffering_strategy: "all", "fixed" or "relative"
                                   "all" buffers the whole matrix
                                   "fixed" buffers a fixed area, specified by buffering_arg
                                           around the query area
                                   "relative" buffers a multiple of the query area.
                                              With buffering_arg=1 the query area plus
                                              the same amount upstream and downstream
                                              are buffered
        :param buffering_arg: Number specifying how much around the query area is buffered
        """
        self.data = data
        if buffering_strategy not in self._BUFFERING_STRATEGIES:
            raise ValueError("Only support the buffering strategies {}".format(list(self._BUFFERING_STRATEGIES.keys())))
        self.buffering_strategy = buffering_strategy
        self.buffering_arg = buffering_arg
        self.buffered_region = None
        self.buffered_matrix = None
        self.weight_field = weight_field
        self.default_value = default_value
        self.smooth_sigma = smooth_sigma
        self.norm = norm
        self.oe = oe
        self.log = log

    @classmethod
    def from_hic_matrix(cls, hic_matrix, weight_field=None, default_value=None,
                        smooth_sigma=None, norm=True, oe=False, log=False):
        """
        Wrap a :class:`~HicMatrix` in a :class:`~BufferedMatrix` container.
        Default buffering strategy is set to "all" by default.

        :param hic_matrix: :class:`~HicMatrix`
        :return: :class:`~BufferedMatrix`
        """
        bm = cls(data=None, buffering_strategy="all", weight_field=weight_field,
                 default_value=default_value, smooth_sigma=smooth_sigma,
                 norm=norm, oe=oe, log=log)
        bm.buffered_region = bm._STRATEGY_ALL
        bm.buffered_matrix = hic_matrix
        return bm

    def is_buffered_region(self, *regions):
        """
        Check if set of :class:`~GenomicRegion`s is already buffered in this matrix.

        :param regions: :class:`~GenomicRegion` object(s)
        :return:
        """
        if (self.buffered_region is None or self.buffered_matrix is None or
                (not self.buffered_region == self._STRATEGY_ALL and not
                 all(rb.contains(rq) for rb, rq in zip(self.buffered_region, regions)))):
            return False
        return True

    def get_matrix(self, *regions):
        """
        Retrieve a sub-matrix by the given :class:`~GenomicRegion` object(s).

        Will automatically load data if a non-buffered region is requested.

        :param regions: :class:`~GenomicRegion` object(s)
        :return: :class:`~HicMatrix`
        """
        # regions = tuple(reversed([r for r in regions]))
        if not self.is_buffered_region(*regions):
            logger.debug("Buffering matrix")
            self._BUFFERING_STRATEGIES[self.buffering_strategy](self, *regions)
        m = self.buffered_matrix[tuple(regions)]
        if self.smooth_sigma is not None:
            mf = gaussian_filter(m, self.smooth_sigma)
            m = fanc.matrix.RegionMatrix(mf, row_regions=m.row_regions, col_regions=m.col_regions)
        return m

    def _buffer_all(self, *regions):
        """
        No buffering, just loads everything in the object into memory.

        Obviously very memory intensive.

        :param regions: :class:`~GenomicRegion` objects
        :return: :class:`~HicMatrix`
        """
        self.buffered_region = self._STRATEGY_ALL
        self.buffered_matrix = self.data.matrix(key=tuple([slice(0, None, None)]*len(regions)),
                                                score_field=self.weight_field,
                                                default_value=self.default_value,
                                                norm=self.norm, oe=self.oe, log=self.log)

    def _buffer_relative(self, *regions):
        """
        Load the requested :class:`~GenomicRegion` and buffer an additional fration
        of the matrix given by buffering_arg*len(region)

        :param regions: :class:`~GenomicRegion` objects
        :return: :class:`~HicMatrix`
        """
        self.buffered_region = []
        for rq in regions:
            if rq.start is not None and rq.end is not None:
                rq_size = rq.end - rq.start
                new_start = max(1, rq.start - rq_size*self.buffering_arg)
                new_end = rq.end + rq_size*self.buffering_arg
                self.buffered_region.append(GenomicRegion(start=new_start, end=new_end,
                                                          chromosome=rq.chromosome))
            else:
                self.buffered_region.append(GenomicRegion(start=None, end=None, chromosome=rq.chromosome))
        self.buffered_matrix = self.data.matrix(tuple(self.buffered_region),
                                                score_field=self.weight_field,
                                                default_value=self.default_value,
                                                norm=self.norm, oe=self.oe, log=self.log)

    def _buffer_fixed(self, *regions):
        """
        Load the requested :class:`~GenomicRegion` and buffer an additional
        fixed part of the matrix given by buffering_arg

        :param regions: :class:`~GenomicRegion` objects
        :return: :class:`~HicMatrix`
        """
        self.buffered_region = []
        for rq in regions:
            if rq.start is not None and rq.end is not None:
                new_start = max(1, rq.start - self.buffering_arg)
                new_end = rq.end + self.buffering_arg
                self.buffered_region.append(GenomicRegion(start=new_start, end=new_end,
                                                          chromosome=rq.chromosome))
            else:
                self.buffered_region.append(GenomicRegion(start=None, end=None, chromosome=rq.chromosome))
        self.buffered_matrix = self.data.matrix(tuple(self.buffered_region),
                                                score_field=self.weight_field,
                                                default_value=self.default_value,
                                                norm=self.norm, oe=self.oe, log=self.log)

    @property
    def buffered_min(self):
        """
        Find the smallest non-zero buffered matrix value.

        :return: float or None if nothing is buffered
        """
        return float(np.nanmin(self.buffered_matrix[np.ma.nonzero(self.buffered_matrix)]))\
            if self.buffered_matrix is not None else None

    @property
    def buffered_max(self):
        """
        Find the largest buffered matrix value
        :return: float or None if nothing is buffered
        """
        return float(np.nanmax(self.buffered_matrix)) if self.buffered_matrix is not None else None

    _BUFFERING_STRATEGIES = {_STRATEGY_ALL: _buffer_all,
                             _STRATEGY_RELATIVE: _buffer_relative,
                             _STRATEGY_FIXED: _buffer_fixed}


class BufferedCombinedMatrix(BufferedMatrix):
    """
    A buffered square matrix where values above and below the diagonal
    come from different matrices.
    """
    def __init__(self, top_matrix, bottom_matrix, scale_matrices=True,
                 **kwargs):
        super(BufferedCombinedMatrix, self).__init__(None, **kwargs)

        if top_matrix.bin_size != bottom_matrix.bin_size:
            raise ValueError("Can only combine matrices with equal binning!")

        scaling_factor = top_matrix.scaling_factor(bottom_matrix) if scale_matrices else 1.
        self.data = SplitMatrix(top_matrix, bottom_matrix, scaling_factor=scaling_factor)


class BasePlotterHic(BasePlotterMatrix):
    """
    Base class for plotting Hi-C data.

    **warning: should always be inherited from first, before the other
    base classed**

    Makes use of matrix buffering by :class:`~BufferedMatrix` internally.
    """

    def __init__(self, hic_data, adjust_range=False, buffering_strategy="relative",
                 buffering_arg=1, weight_field=None, default_value=None, smooth_sigma=None,
                 matrix_norm=True, oe=False, log=False, **kwargs):
        """
        :param hic_data: Path to Hi-C data on disk or
                        :class:`~fanc.data.genomic.Hic` or :class:`~fanc.data.genomic.RegionMatrix`
        :param adjust_range: Draw a slider to adjust vmin/vmax interactively. Default: False
        :param buffering_strategy: A valid buffering strategy for :class:`~BufferedMatrix`
        :param buffering_arg: Adjust range of buffering for :class:`~BufferedMatrix`
        """
        super(BasePlotterHic, self).__init__(**kwargs)
        if isinstance(hic_data, string_types):
            hic_data = fanc.load(hic_data, mode="r")
        self.hic_data = hic_data
        self.hic_buffer = prepare_hic_buffer(hic_data, buffering_strategy=buffering_strategy,
                                             buffering_arg=buffering_arg, weight_field=weight_field,
                                             default_value=default_value, smooth_sigma=smooth_sigma,
                                             norm=matrix_norm, oe=oe, log=log)
        self.slider = None
        self.adjust_range = adjust_range
        self.vmax_slider = None


class SquareMatrixPlot(BasePlotterHic, BasePlotter2D):
    """
    Plot Hi-C map as a square matrix.
    """
    sharey = True

    def __init__(self, hic_data, flip=False, **kwargs):
        """
        :param flip: Transpose matrix before plotting
        """
        super(SquareMatrixPlot, self).__init__(hic_data=hic_data, **kwargs)
        self.vmax_slider = None
        self.current_matrix = None
        self.flip = flip
        self.im = None

    def _plot(self, region):
        self._refresh(region)

        self.ax.spines['right'].set_visible(False)
        self.ax.spines['top'].set_visible(False)
        self.ax.xaxis.set_ticks_position('bottom')
        self.ax.yaxis.set_ticks_position('left')

        if self.adjust_range:
            self.add_adj_slider()

        return self.ax

    def add_adj_slider(self, ax=None):
        if ax is None:
            pad = self._total_padding if self._total_padding is not None else 0.
            ax = append_axes(self.ax, 'bottom', config.adjustment_slider_height, pad)

        vmin = self.hic_buffer.buffered_min
        vmax = self.hic_buffer.buffered_max
        self.vmax_slider = Slider(ax, 'vmax', vmin,
                                  vmax, valinit=self.vmax,
                                  facecolor='#dddddd', edgecolor='none')

        self.vmax_slider.on_changed(self._slider_refresh)

        self.ax.patch.set_visible(False)

    def _slider_refresh(self, val):
        # new_vmin = self.vmin_slider.val
        new_vmax = self.vmax_slider.val
        if self.colorbar_symmetry is not None:
            diff = abs(self.colorbar_symmetry - new_vmax)
            self.im.set_clim(vmin=self.colorbar_symmetry - diff, vmax=self.colorbar_symmetry + diff)
        else:
            self.im.set_clim(vmin=self.hic_buffer.buffered_min, vmax=new_vmax)
        # Hack to force redraw of image data
        self._refresh(None)

        if self.colorbar is not None:
            self.update_colorbar(vmax=new_vmax)

    def _refresh(self, region):
        if region is not None:
            self.current_matrix = self.hic_buffer.get_matrix(*region)
        old_image = self.im

        m = np.transpose(self.current_matrix)
        if self.flip:
            m = np.flipud(m)
            extent = [self.current_matrix.row_regions[0].start, self.current_matrix.row_regions[-1].end,
                      self.current_matrix.col_regions[0].start, self.current_matrix.col_regions[-1].end]
        else:
            extent = [self.current_matrix.row_regions[0].start, self.current_matrix.row_regions[-1].end,
                      self.current_matrix.col_regions[-1].end, self.current_matrix.col_regions[0].start]

        color_matrix = self.get_color_matrix(m)
        self.im = self.ax.imshow(color_matrix, interpolation='none', aspect='auto',
                                 cmap=self.colormap, norm=self._map_norm, extent=extent)

        self.ax.set_xlim(extent[0], extent[1])
        self.ax.set_ylim(extent[2], extent[3])

        if old_image is not None:
            old_image.remove()


HicPlot2D = SquareMatrixPlot


class SplitMatrixPlot(HicPlot2D):
    def __init__(self, hic_top, hic_bottom, scale_matrices=True,
                 weight_field=None, default_value=None, smooth_sigma=None,
                 matrix_norm=True, oe=False, log=False, **kwargs):
        super(SplitMatrixPlot, self).__init__(hic_top, **kwargs)
        self.hic_top = hic_top
        self.hic_bottom = hic_bottom
        self.hic_buffer = BufferedCombinedMatrix(hic_bottom, hic_top, scale_matrices=scale_matrices,
                                                 weight_field=weight_field, default_value=default_value,
                                                 smooth_sigma=smooth_sigma, norm=matrix_norm, oe=oe,
                                                 log=log)


HicComparisonPlot2D = SplitMatrixPlot


class HicSlicePlot(ScalarDataPlot):
    """
    Draw Hi-C data as virtual 4C-plot. All interactions that
    involve the slice region are shown.
    """

    def __init__(self, hic_data, slice_region, names=None,
                 colors=None, fill=None,
                 buffering_strategy="relative", buffering_arg=1,
                 weight_field=None, default_value=None, **kwargs):
        """
        :param hic_data: :class:`~fanc.Hic` or :class:`~fanc.RegionMatrix`. Can be list of
                         multiple Hi-C datasets.
        :param slice_region: String ("2L:1000000-1500000") or :class:`~GenomicRegion`.
                             All interactions involving this region are shown.
        :param names: If multiple Hi-C datasets are provided, can pass a list of names.
                      Are used as names in the legend of the plot.
        :param buffering_strategy: A valid buffering strategy for class:`~BufferedMatrix`
        :param buffering_arg: Adjust range of buffering for class:`~BufferedMatrix`
        """
        kwargs.setdefault("aspect", .3)
        super(HicSlicePlot, self).__init__(**kwargs)
        if not isinstance(hic_data, (list, tuple)):
            hic_data = [hic_data]
        self.hic_buffers = []
        for h in hic_data:
            hb = prepare_hic_buffer(h,
                                    buffering_strategy=buffering_strategy,
                                    buffering_arg=buffering_arg,
                                    weight_field=weight_field,
                                    default_value=default_value)
            self.hic_buffers.append(hb)
        self.names = names
        if isinstance(slice_region, string_types):
            slice_region = GenomicRegion.from_string(slice_region)
        self.slice_region = slice_region
        self.x = None
        self.y = None
        self.lines = []
        self.fill = fill

        if colors is None:
            prop_cycle = plt.rcParams['axes.prop_cycle']
            colors = prop_cycle.by_key()['color']
        elif isinstance(colors, string_types):
            colors = [colors]

        self.colors = colors

    def _refresh(self, region):
        for line in self.lines:
            line.remove()
        self.lines = []

        color_cycle = cycle(self.colors)
        for i, b in enumerate(self.hic_buffers):
            hm = b.get_matrix(self.slice_region, region.expand(relative=0.5))
            m = np.mean(hm, axis=0)
            bin_coords = np.r_[[x.start for x in hm.col_regions], hm.col_regions[-1].end]
            bin_coords = (bin_coords[1:] + bin_coords[:-1])/2
            line = self.ax.plot(bin_coords, m,
                                color=next(color_cycle),
                                label=self.names[i] if self.names else "")[0]
            self.lines.append(line)

            if self.fill:
                self.ax.fill_between(bin_coords, [0] * len(m), m, color=line.get_color())

    def _plot(self, region):
        self._refresh(region)

        # plot patch at slice region
        bottom, top = self.ax.get_ylim()
        patch_region = self.slice_region.copy()

        rect = patches.Rectangle((patch_region.start, bottom),
                                 len(patch_region), (top - bottom)/20,
                                 facecolor='grey', edgecolor='grey')
        self.ax.add_patch(rect)

        if self.names:
            self.add_legend()
        self.remove_colorbar_ax()
        sns.despine(ax=self.ax, top=True, right=True)


class TriangularMatrixPlot(BasePlotterHic, BasePlotter1D):
    """
    A triangle Hi-C heatmap plot.
    """

    def __init__(self, hic_data, max_dist=None, proportional=True,
                 rasterized=True, **kwargs):
        """
        :param max_dist: Only draw interactions up to this distance
        :param proportional: Automatically determine aspect ratio of plot
                             so that x- and y-axis are proportional. Default: True
        :param rasterized: Draw map as image (True) or vector graphic (False).
                           Default: True
        """
        kwargs.setdefault("aspect", None)
        super(TriangularMatrixPlot, self).__init__(hic_data=hic_data, **kwargs)
        self.proportional = proportional
        self.max_dist = str_to_int(max_dist)
        self.hm = None
        self.rasterized = rasterized
        self.collection = None

    def _plot(self, region):
        logger.debug("Generating matrix from hic object")
        if region is None:
            raise ValueError("Cannot plot triangle plot for whole genome.")
        if region.start is None:
            region.start = 1
        if region.end is None:
            region.end = self.hic_data.chromosome_lengths[region.chromosome]
        if self.aspect is None and self.proportional:
            if self.max_dist is None:
                self.aspect = .5
            else:
                rl = region.end - region.start
                self.aspect = .5*min(self.max_dist, rl)/rl
            self._dimensions_stale = True

        # Have to copy unfortunately, otherwise modifying matrix in buffer
        self._refresh(region)

        # set limits and aspect ratio
        # self.ax.set_aspect(aspect="equal")
        self.ax.set_ylim(0, self.max_dist/2 if self.max_dist else (region.end-region.start)/2)
        # remove outline everywhere except at bottom
        sns.despine(ax=self.ax, top=True, right=True, left=True)
        self.ax.set_yticks([])
        # hide background patch
        self.ax.patch.set_visible(False)
        if self.adjust_range:
            self.add_adj_slider()

        def drag_pan(self, button, key, x, y):
            mpl.axes.Axes.drag_pan(self, button, 'x', x, y)  # pretend key=='x'

        self.ax.drag_pan = types.MethodType(drag_pan, self.ax)

    def _mesh_data(self, region):
        hm = self.hic_buffer.get_matrix(region, region)
        hm_copy = fanc.matrix.RegionMatrix(np.copy(hm), col_regions=hm.col_regions,
                                           row_regions=hm.row_regions)
        # update coordinates
        bin_coords = np.r_[[x.start for x in hm_copy.row_regions], hm_copy.row_regions[-1].end]
        # Make sure the matrix is not protruding over the end of the requested plotting region
        if bin_coords[0] < region.start <= bin_coords[1]:
            bin_coords[0] = region.start
        if bin_coords[-1] > region.end >= bin_coords[-2]:
            bin_coords[-1] = region.end
        bin_coords = np.true_divide(bin_coords, np.sqrt(2))
        x, y = np.meshgrid(bin_coords, bin_coords)
        # rotatate coordinate matrix 45 degrees
        sin45 = np.sin(np.radians(45))
        x_, y_ = x * sin45 + y * sin45, x * sin45 - y * sin45

        return x_, y_, hm_copy

    def _update_mesh_colors(self):
        # pcolormesh doesn't support plotting RGB arrays directly like imshow, have to workaround
        # See https://github.com/matplotlib/matplotlib/issues/4277
        # http://stackoverflow.com/questions/29232439/plotting-an-irregularly-spaced-rgb-image-in-python/29232668?noredirect=1#comment46710586_29232668
        color_matrix = self.get_color_matrix(self.hm)
        color_tuple = color_matrix.transpose((1, 0, 2)).reshape(
            (color_matrix.shape[0] * color_matrix.shape[1], color_matrix.shape[2]))
        self.collection.set_facecolor(color_tuple)

    def _refresh(self, region):
        x_, y_, hm = self._mesh_data(region)
        self.hm = hm

        old_collection = self.collection
        self.collection = self.ax.pcolormesh(x_, y_, hm, cmap=self.colormap, norm=self._map_norm,
                                             rasterized=self.rasterized)
        self.collection._A = None
        self._update_mesh_colors()

        if old_collection is not None:
            old_collection.remove()

    def add_adj_slider(self, ax=None):
        if ax is None:
            pad = self._total_padding if self._total_padding is not None else 0.
            ax = append_axes(self.ax, 'bottom', config.adjustment_slider_height, pad)

        self.vmax_slider = Slider(ax, 'vmax', self.hic_buffer.buffered_min,
                                  self.hic_buffer.buffered_max, valinit=self.vmax,
                                  facecolor='#dddddd', edgecolor='none')

        self.vmax_slider.on_changed(self._slider_refresh)

    def _slider_refresh(self, val):
        # new_vmin = self.vmin_slider.val
        new_vmax = self.vmax_slider.val
        if self.colorbar_symmetry is not None:
            diff = abs(self.colorbar_symmetry-new_vmax)
            self._update_norm(vmin=self.colorbar_symmetry-diff, vmax=self.colorbar_symmetry+diff)
        else:
            self._update_norm(vmax=new_vmax)
        self._update_mesh_colors()
        if self.colorbar is not None:
            self.update_colorbar(vmax=new_vmax)


HicPlot = TriangularMatrixPlot


class HicPeakPlot(BaseOverlayPlotter):
    """
    Overlay peaks onto Hicplot or HicPlot2D. Accepts
    :class:`~fanc.data.network.PeakInfo`.
    Add to HicPlot or HicPlot2D using add_overlay method.
    """
    def __init__(self, peaks, radius=None, circle_props={}, **kwargs):
        """
        :param peaks: fanc peaks instance
        :param radius: Radius in bp for plotted circles.
                       If not specified (default), use the radius of the
                       peak itself. This is often too small to see,
                       providing a value like 50000 helps. Default: None
        :param circe_props: Dictionary with properties for the plotted circles
                            for the matplotlib.patches.Circle constructor.
                            Default: Black edges, no fill, linewidth 3 pt
        """
        super(HicPeakPlot, self).__init__(**kwargs)
        self.peaks = peaks
        self.radius = radius
        self.circle_props = {
            "edgecolor": "black",
            "fill": False,
            "linewidth": 3,
        }
        self.circle_props.update(circle_props)

    @property
    def compatibility(self):
        return ["HicPlot", "HicPlot2D"]

    def _plot(self, base_plot, region):
        def plot_hicplot(start, end, radius):
            x = .5*(start + end)
            y = .5*(end - start)
            circle = patches.Circle((x, y), radius, **self.circle_props)
            base_plot.ax.add_patch(circle)

        def plot_hicplot2d(start, end, radius):
            circle = patches.Circle((start, end), radius, **self.circle_props)
            base_plot.ax.add_patch(circle)
            circle = patches.Circle((end, start), radius, **self.circle_props)
            base_plot.ax.add_patch(circle)

        plot_dispatch = {
            "HicPlot": plot_hicplot,
            "HicPlot2D": plot_hicplot2d
        }
        base_plot_class = base_plot.__class__.__name__
        plot_func = plot_dispatch[base_plot_class]
        peaks_gen = self.peaks.edge_subset((region, region), distances_in_bp=True)
        for p in peaks_gen:
            plot_func((p.source_node.start + p.source_node.end)/2,
                      (p.sink_node.start + p.sink_node.end)/2,
                      self.radius if self.radius is not None else getattr(p, "radius", self.peaks.bin_size*3)*self.peaks.bin_size)


class EdgeFilterBuffer(object):
    def __init__(self, hic, plot_field='weight', default_value=0, log2=False):
        self.hic = hic
        self._region_buffer = dict()
        self._edge_buffer = dict()
        self._filters = []
        self._plot_field = plot_field
        self._default_value = default_value
        self._log2 = log2
        self._last_matrix = None

    def add_filter(self, filter):
        self._filters.append(filter)

    @property
    def buffered_min(self):
        """
        Find the smallest non-zero buffered matrix value.

        :return: float or None if nothing is buffered
        """
        return float(np.nanmin(self._last_matrix[np.ma.nonzero(self._last_matrix)])) \
            if self._last_matrix is not None else None

    @property
    def buffered_max(self):
        """
        Find the largest buffered matrix value
        :return: float or None if nothing is buffered
        """
        return float(np.nanmax(self._last_matrix)) if self._last_matrix is not None else None

    def get_matrix(self, *regions):
        row_region = regions[0]
        col_region = regions[1]
        row_key = (row_region.chromosome, row_region.start, row_region.end)
        col_key = (col_region.chromosome, col_region.start, col_region.end)

        if row_key in self._region_buffer:
            sub_row_regions = self._region_buffer[row_key]
        else:
            sub_row_regions = list(self.hic.regions(row_region))
            self._region_buffer[row_key] = sub_row_regions

        if col_key in self._region_buffer:
            sub_col_regions = self._region_buffer[col_key]
        else:
            sub_col_regions = list(self.hic.regions(col_region))
            self._region_buffer[col_key] = sub_col_regions

        row_offset = sub_row_regions[0].ix
        col_offset = sub_col_regions[0].ix

        if (row_key, col_key) in self._edge_buffer:
            edges = self._edge_buffer[(row_key, col_key)]
        else:
            edges = list(self.hic.edge_subset((row_region, col_region)))
            self._edge_buffer[(row_key, col_key)] = edges

        m = np.full((len(sub_row_regions), len(sub_col_regions)), self._default_value,
                    dtype=np.float64)

        n_filtered = 0
        n_filtered_by_filter = defaultdict(int)
        for edge in edges:
            valid = True
            for filter_ix, filter in enumerate(self._filters):
                if not filter.valid_peak(edge):
                    n_filtered_by_filter[filter_ix] += 1
                    valid = False
                    break

            if not valid:
                n_filtered += 1
                continue

            source = edge.source
            sink = edge.sink
            weight = getattr(edge, self._plot_field)
            if self._log2:
                weight = np.log2(weight)

            ir = source - row_offset
            jr = sink - col_offset
            if 0 <= ir < m.shape[0] and 0 <= jr < m.shape[1]:
                m[ir, jr] = weight

            ir = sink - row_offset
            jr = source - col_offset
            if 0 <= ir < m.shape[0] and 0 <= jr < m.shape[1]:
                m[ir, jr] = weight

        rm = RegionMatrix(m, row_regions=sub_row_regions, col_regions=sub_col_regions)
        self._last_matrix = rm
        return rm


class EdgeHicPlot(HicPlot2D):
    def __init__(self, hic, plot_field='weight', default_value=0,
                 log2=False, highlight_edges=False,
                 highlight_limit=100, *args, **kwargs):
        HicPlot2D.__init__(self, hic, *args, **kwargs)

        self.hic_buffer = EdgeFilterBuffer(hic, plot_field=plot_field,
                                           default_value=default_value,
                                           log2=log2)
        self._highlight_edges = highlight_edges
        self._highlight_circles = []
        self._highlight_limit = highlight_limit

    def update_highlights(self, m=None):
        for circle in self._highlight_circles:
            circle.remove()

        self._highlight_circles = []

        if self._highlight_edges and m is not None:
            all_x, all_y = np.where(m > 0)
            if not len(all_x) > self._highlight_limit:
                for x, y in zip(all_x, all_y):
                    region_x = m.row_regions[x]
                    region_y = m.col_regions[y]

                    circle = patches.Circle((region_x.center, region_y.center), len(region_x)*3,
                                            facecolor='red')
                    self.ax.add_patch(circle)
                    self._highlight_circles.append(circle)

    def _plot(self, region):
        HicPlot2D._plot(self, region)

    def _refresh(self, region):
        HicPlot2D._refresh(self, region)


class PeakParameterPlot(object):
    def __init__(self, peaks, font_size=5, observed_init=1.,
                 oe_d_init=1., oe_h_init=1., oe_v_init=1., oe_l_init=1.,
                 fdr_d_init=.1, fdr_h_init=.1, fdr_v_init=.1, fdr_l_init=.1,
                 mappability_d_init=0., mappability_h_init=0.,
                 mappability_v_init=0., mappability_l_init=0.,
                 oe_slider_range=(0, 5), oe_slider_step=0.1,
                 fdr_slider_range=(0, .1), fdr_slider_step=0.001,
                 mappability_slider_range=(0, 1.), mappability_slider_step=0.05,
                 observed_range=(0, 50), observed_step=1,
                 **kwargs):
        self.peaks = peaks
        self.font_size = font_size
        self.hic_args = kwargs
        
        self.observed_init = observed_init

        self.oe_init = {
            'd': oe_d_init,
            'v': oe_v_init,
            'h': oe_h_init,
            'l': oe_l_init,
        }

        self.fdr_init = {
            'd': fdr_d_init,
            'v': fdr_v_init,
            'h': fdr_h_init,
            'l': fdr_l_init,
        }

        self.mappability_init = {
            'd': mappability_d_init,
            'v': mappability_v_init,
            'h': mappability_h_init,
            'l': mappability_l_init,
        }

        self.fdr_range = fdr_slider_range
        self.fdr_step = fdr_slider_step
        self.oe_range = oe_slider_range
        self.oe_step = oe_slider_step
        self.mappability_range = mappability_slider_range
        self.mappability_step = mappability_slider_step
        self.observed_range = observed_range
        self.observed_step = observed_step

        self.observed_cutoff = self.observed_init
        self.oe_cutoffs = self.oe_init.copy()
        self.fdr_cutoffs = self.fdr_init.copy()
        self.mappability_cutoffs = self.mappability_init.copy()

        self.ax_fdr_sliders = {}
        self.fdr_sliders = {}
        self.oe_sliders = {}
        self.ax_oe_sliders = {}
        self.ax_mappability_sliders = {}
        self.mappability_sliders = {}
        self.observed_slider = None

        self.observed_filter = None
        self.fdr_filter = None
        self.mappability_filter = None
        self.oe_filter = None

        self.filtered_plots = []
        self.hic_plots = []

        self.button = None
        self.region_pairs = []

        self.fig = None

    def plot(self, *regions):
        plt.rcParams.update({'font.size': self.font_size})

        # process region pairs
        self.region_pairs = []
        for pair in regions:
            if isinstance(pair, string_types):
                r = as_region(pair)
                pair = (r, r)
            elif isinstance(pair, GenomicRegion):
                pair = (pair, pair)

            try:
                r1, r2, vmax = pair
            except ValueError:
                r1, r2 = pair
                if isinstance(r2, float) or isinstance(r2, int):
                    vmax = r2
                    r2 = r1
                else:
                    vmax = None
            r1 = as_region(r1)
            r2 = as_region(r2)
            self.region_pairs.append((r1, r2, vmax))

        #
        # necessary plots: hic, oe_d, fdr_d, uncorrected,
        #
        gs = grd.GridSpec(len(self.region_pairs) + 4, 4,
                          height_ratios=[10] * len(self.region_pairs) + [1, 1, 1, 3],
                          wspace=0.3, hspace=0.5)

        self.fig = plt.figure(figsize=(10, len(self.region_pairs) * 2 + 2), dpi=150)

        # sliders
        inner_observed_gs = grd.GridSpecFromSubplotSpec(3, 1,
                                                        subplot_spec=gs[len(self.region_pairs) + 3, 0],
                                                        wspace=0.0, hspace=0.0)

        ax_observed_slider = plt.subplot(inner_observed_gs[0, 0])
        self.observed_slider = Slider(ax_observed_slider, 'uncorrected',
                                      self.observed_range[0], self.observed_range[1],
                                      valinit=self.observed_init, valstep=self.observed_step)

        self.ax_oe_sliders = dict()
        self.oe_sliders = dict()
        self.ax_fdr_sliders = dict()
        self.fdr_sliders = dict()
        self.ax_mappability_sliders = dict()
        self.mappability_sliders = dict()
        for i, neighborhood in enumerate(['d', 'h', 'v', 'l']):
            # O/E
            self.ax_oe_sliders[neighborhood] = plt.subplot(gs[len(self.region_pairs) + 0, i])
            self.oe_sliders[neighborhood] = Slider(self.ax_oe_sliders[neighborhood],
                                                   'O/E {}'.format(neighborhood.upper()),
                                                   self.oe_range[0], self.oe_range[1],
                                                   valinit=self.oe_init[neighborhood],
                                                   valstep=self.oe_step)
            # FDR
            self.ax_fdr_sliders[neighborhood] = plt.subplot(gs[len(self.region_pairs) + 1, i])
            self.fdr_sliders[neighborhood] = Slider(self.ax_fdr_sliders[neighborhood],
                                                    'FDR {}'.format(neighborhood.upper()),
                                                    self.fdr_range[0], self.fdr_range[1],
                                                    valinit=self.fdr_init[neighborhood],
                                                    valstep=self.fdr_step)
            # Mappability
            self.ax_mappability_sliders[neighborhood] = plt.subplot(gs[len(self.region_pairs) + 2, i])
            self.mappability_sliders[neighborhood] = Slider(self.ax_mappability_sliders[neighborhood],
                                                            'Map {}'.format(neighborhood.upper()),
                                                            self.mappability_range[0],
                                                            self.mappability_range[1],
                                                            valinit=self.mappability_init[neighborhood],
                                                            valstep=self.mappability_step)

        # check button
        inner_button_gs = grd.GridSpecFromSubplotSpec(1, 2, subplot_spec=gs[len(self.region_pairs) + 3, 1],
                                                      wspace=0.0, hspace=0.0)
        ax_button = plt.subplot(inner_button_gs[0, 0])
        self.button = CheckButtons(ax_button, ['Show loops'], [False])

        # filters
        self.observed_filter = ObservedPeakFilter(cutoff=self.observed_init)
        self.oe_filter = EnrichmentPeakFilter(enrichment_d_cutoff=self.oe_init['d'],
                                              enrichment_h_cutoff=self.oe_init['h'],
                                              enrichment_v_cutoff=self.oe_init['v'],
                                              enrichment_ll_cutoff=self.oe_init['l'])
        self.fdr_filter = FdrPeakFilter(fdr_d_cutoff=self.fdr_init['d'],
                                        fdr_ll_cutoff=self.fdr_init['h'],
                                        fdr_h_cutoff=self.fdr_init['v'],
                                        fdr_v_cutoff=self.fdr_init['l'])
        self.mappability_filter = MappabilityPeakFilter(mappability_d_cutoff=self.mappability_init['d'],
                                                        mappability_h_cutoff=self.mappability_init['h'],
                                                        mappability_v_cutoff=self.mappability_init['v'],
                                                        mappability_ll_cutoff=self.mappability_init['l'])

        self.hic_plots = []
        self.filtered_plots = []
        for i, (r1, r2, vmax) in enumerate(self.region_pairs):
            ax_hic = plt.subplot(gs[i, 0])
            ax_filtered = plt.subplot(gs[i, 1])
            ax_oe = plt.subplot(gs[i, 2])
            ax_fdr = plt.subplot(gs[i, 3])

            hic_plot = EdgeHicPlot(self.peaks,
                                   ax=ax_hic, show_colorbar=False, adjust_range=False,
                                   unmappable_color='white', vmax=vmax,
                                   highlight_edges=True, highlight_limit=200,
                                   draw_tick_legend=False,
                                   **self.hic_args)
            filtered_plot = EdgeHicPlot(self.peaks,
                                        ax=ax_filtered, show_colorbar=False, adjust_range=False,
                                        unmappable_color='white', vmax=vmax,
                                        draw_tick_legend=False,
                                        **self.hic_args)
            oe_plot = EdgeHicPlot(self.peaks, plot_field='oe_d', default_value=0, norm='lin',
                                  ax=ax_oe, show_colorbar=False, colormap='white_red', adjust_range=False,
                                  vmin=1, vmax=4, log2=False, unmappable_color='white',
                                  draw_tick_legend=False,)
            fdr_plot = EdgeHicPlot(self.peaks, plot_field='fdr_d', default_value=1, norm='lin',
                                   ax=ax_fdr, show_colorbar=False, adjust_range=False,
                                   vmin=0, vmax=0.05, colormap='Greys_r', unmappable_color='white',
                                   draw_tick_legend=False)

            filtered_plot.hic_buffer.add_filter(self.observed_filter)
            filtered_plot.hic_buffer.add_filter(self.oe_filter)
            filtered_plot.hic_buffer.add_filter(self.fdr_filter)
            filtered_plot.hic_buffer.add_filter(self.mappability_filter)

            self.hic_plots.append(hic_plot)
            self.filtered_plots.append(filtered_plot)

            hic_plot.plot((r1, r2))
            filtered_plot.plot((r1, r2))
            oe_plot.plot((r1, r2))
            fdr_plot.plot((r1, r2))

            ax_filtered.set_yticklabels([])
            ax_oe.set_yticklabels([])
            ax_fdr.set_yticklabels([])

        for neighborhood in ['d', 'h', 'v', 'l']:
            self.oe_sliders[neighborhood].on_changed(self.update_oe_filter)
            self.fdr_sliders[neighborhood].on_changed(self.update_fdr_filter)
            self.mappability_sliders[neighborhood].on_changed(self.update_mappability_filter)

        self.observed_slider.on_changed(self.update_observed_filter)

        self.button.on_clicked(self.refresh_plots)
        self.refresh_plots()
        return self.fig

    def refresh_plots(self, event=None):
        for i, (r1, r2, vmax) in enumerate(self.region_pairs):
            self.filtered_plots[i].refresh((r1, r2))

            if self.button.get_status()[0]:
                self.hic_plots[i].update_highlights(self.filtered_plots[i].hic_buffer._last_matrix)
            else:
                self.hic_plots[i].update_highlights()

        self.fig.canvas.draw()

    def update_observed_filter(self, event):
        self.observed_cutoff = self.observed_slider.val
        self.observed_filter.cutoff = self.observed_cutoff
        logger.info("Observed cutoff set to {}".format(self.observed_cutoff))
        self.refresh_plots()

    def update_oe_filter(self, event):
        self.oe_cutoffs = {
            'd': self.oe_sliders['d'].val,
            'h': self.oe_sliders['h'].val,
            'v': self.oe_sliders['v'].val,
            'l': self.oe_sliders['l'].val
        }

        self.oe_filter.enrichment_d_cutoff = self.oe_cutoffs['d']
        self.oe_filter.enrichment_h_cutoff = self.oe_cutoffs['h']
        self.oe_filter.enrichment_v_cutoff = self.oe_cutoffs['v']
        self.oe_filter.enrichment_ll_cutoff = self.oe_cutoffs['l']
        logger.info("O/E cutoffs set to: {}".format(self.oe_cutoffs))
        self.refresh_plots()

    def update_fdr_filter(self, event):
        self.fdr_cutoffs = {
            'd': self.fdr_sliders['d'].val,
            'h': self.fdr_sliders['h'].val,
            'v': self.fdr_sliders['v'].val,
            'l': self.fdr_sliders['l'].val
        }
        self.fdr_filter.fdr_d_cutoff = self.fdr_cutoffs['d']
        self.fdr_filter.fdr_h_cutoff = self.fdr_cutoffs['h']
        self.fdr_filter.fdr_v_cutoff = self.fdr_cutoffs['v']
        self.fdr_filter.fdr_ll_cutoff = self.fdr_cutoffs['l']

        logger.info("FDR cutoffs set to: {}".format(self.fdr_cutoffs))
        self.refresh_plots()

    def update_mappability_filter(self, event):
        self.mappability_cutoffs = {
            'd': self.mappability_sliders['d'].val,
            'h': self.mappability_sliders['h'].val,
            'v': self.mappability_sliders['v'].val,
            'l': self.mappability_sliders['l'].val
        }
        self.mappability_filter.mappability_d_cutoff = self.mappability_cutoffs['d']
        self.mappability_filter.mappability_h_cutoff = self.mappability_cutoffs['h']
        self.mappability_filter.mappability_v_cutoff = self.mappability_cutoffs['v']
        self.mappability_filter.mappability_ll_cutoff = self.mappability_cutoffs['l']

        logger.info("Mappability cutoffs set to: {}".format(self.mappability_cutoffs))
        self.refresh_plots()
